#include "CRingBuffer.h"

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cmath>
#include <string>

using namespace std;

CRingBuffer::CRingBuffer()
	:
	_bufferL( NULL ),
	_bufferR( NULL ),
	_pos( 0 ),
	_size( 0 )
{
	cout << "CRingBuffer() default constructor. Don't forget to call 'init'. " << endl;
}

CRingBuffer::CRingBuffer( unsigned int bufferSize, unsigned int samplerate, float msFadeout )
{
	CRingBuffer::init( bufferSize, samplerate, msFadeout);
}

void CRingBuffer::init(unsigned int ringbufSize, unsigned int sampleRate, float msFadeOut )
{
	_bufferL = NULL;
	_bufferR = NULL;
	_pos = 0;

	_fullCycleFilled = false;
	_filledWithXSamples = 0;

	_sampleRate = sampleRate;
	_msFadeout = msFadeOut;

	setSize( ringbufSize );
}

void CRingBuffer::setSize(unsigned int newSize)//size in samples.
{
	deleteBuffer();// If there are buffers. delete them.
	_pos = 0;
	_size = newSize;
	_fullCycleFilled = false;
	_filledWithXSamples = 0;

	// Dynamically allocate an array of floats. '()'at the end sets all elements to 0.f;
	_bufferL = new float[_size]();
	_bufferR = new float[_size]();
}

void CRingBuffer::deleteBuffer()
{
    if( _bufferL && _bufferR )
    {
        delete[] _bufferL;
        delete[] _bufferR;
		
		_bufferL = NULL;
		_bufferR = NULL;
    }
}

void CRingBuffer::clearBuffer()
{
	for (unsigned int i = 0; i < _size; i++)
	{
		_bufferL[i] = 0.f;
		_bufferR[i] = 0.f;
	}
}

void CRingBuffer::reset()
{
	// Check if _bufferL and _bufferR point to something
	assert( _bufferL && _bufferR );
   
	// Set all elements in the ringbuffer to 0.f;
	CRingBuffer::clearBuffer();
	
	_fullCycleFilled = false;
	_filledWithXSamples = 0;
	_pos = 0;
	return;
}


void CRingBuffer::processFill(float * L, float * R, unsigned int blockSize)
{
	assert( blockSize <= _size );// There must be enough space in the '_ringbuffer' for one block of samples.
    
	_filledWithXSamples += blockSize;
	if( _filledWithXSamples >= _size ) _fullCycleFilled = true;
	
	// cout << "_pos+bufSize:\t" << _pos+bufSize << endl;

    // Check if we can copy the whole 'L'and'R' buffer to the ringbuf without having to wrap. 
	if((_pos + blockSize) < _size )
    {
		int numBytes = sizeof(float) * blockSize;
		memcpy(&_bufferL[_pos], L, numBytes);
		memcpy(&_bufferR[_pos], R, numBytes);
		_pos += blockSize;
	}
    else// In case we have to wrap.
    {
		// Copy the samples that fit in the ringbuf from '_pos' till '_size'.
        int numBytes = sizeof(float) * (_size - _pos);
		memcpy(&_bufferL[_pos], L, numBytes);
		memcpy(&_bufferR[_pos], R, numBytes);
		
		// 'rest' is the number of samples that are not written to the buffer yet. 
		// ( And have to be written to the beginning of the ringbuf)
		int rest = _pos + blockSize - _size;
        // cout << "rest:\t" << rest << endl;
		
		// Copy the rest of the samples to te beginning of the ringbuf.
		numBytes = sizeof(float) * rest;
		memcpy(_bufferL, &L[_size-_pos], numBytes);
		memcpy(_bufferR, &R[_size-_pos], numBytes);
		_pos = rest;
	}
}

void CRingBuffer::reverseChannel_L()
{
	float * temp = new float[_size];
	for (int i = _size - 1; i >= 0; i--)
	{
		temp[i] = _bufferL[i];
	}
	delete[] _bufferL;
	_bufferL = temp;
}

void CRingBuffer::reverseChannel_R()
{
	float * temp = new float[_size];
	for (int i = _size - 1; i >= 0; i--)
	{
		temp[i] = _bufferR[i];
	}
	delete[] _bufferR;
	_bufferR = temp;
}


// Get one linear interpolated sample, for each channel.
// 'index' should be 0.f, '_size' or something inbetween.
// For the purpose of interpolation, there  are two extra elements at the end of '_bufferL' and '_bufferR'.

// _buffer[ _sizeMinus1 ] == _buffer[ _size]; these 2 samples should be te same. 
void CRingBuffer::getSample( float index, float & L, float & R )
{
	assert( index <= (float)_size && index >= 0.f );
	
	// index for the first sample.
	int i = (int) index;

	// index for the second sample.
	int j = i + 1;

	// Extract the part after the floating-point. (fractional part).
	// 'i' is the integral part of 'index'. (an int)
	// So 'f' is like 'index' with the integral part removed.
	float f = index - (float)i;
	
	// Calculate difference between the two samples.
	float difL = _bufferL[j] - _bufferL[i];
    float difR = _bufferR[j] - _bufferR[i];

    L =  _bufferL[i] + ( difL * f );
    R =  _bufferR[i] + ( difR * f );
}

/*
'indexes[]' is an array of indexes with size 'num_samples'
Assumes two channel outputBuf array. ( 1st dimension being channel, 2nd being samples of that channel.)
'getSample()' uses lineair interpolation.
*/
void CRingBuffer::processScratch( float indexes[], unsigned int blockSize, float ** outputBuf )
{
	for ( unsigned int i = 0; i < blockSize; i++ )
	{
		CRingBuffer::getSample( indexes[i], outputBuf[0][i], outputBuf[1][i] );
	}
}

void CRingBuffer::printBufferContent( string text ) const
{
	cout << endl << "ringbuffer content:" << text << endl;
	cout << "_pos is currently:\t" << _pos << endl;
	cout << "The samples and their indexes.:" << endl;
	cout << fixed << setprecision(4);
	for(unsigned int i = 0; i < _size; i++ )
    {
		cout  << "index:\t" << i << "\tcontent L: " << _bufferL[i] <<"\tcontent R: "<< _bufferR[i] << endl;
    }
	cout << "This was ringbuffer with print text:\t" << text << endl;
}

void CRingBuffer::prepareBufferForScratching()
{
	CRingBuffer::applyFadeout();
	// CRingBuffer::printBufferContent("after applyFadeout()");
	CRingBuffer::unwindBuffer();
	// Reversing for one (or both) of the channel(s) could be done here.
}

//'_fullCycleFilled' is true when the 'ringBuffer' contains a full cycle of samples.
// This is intended for finding out, if a fade-out has to be applied in the buffer.
// If no fade-out is applied (in case of a partly filled ringbuffer) we could be scratching over a click/glitch.
void CRingBuffer::applyFadeout()
{
	if (_fullCycleFilled == false)
	{
		assert(_msFadeout > 0.f); // Avoid division with 0.f.
		
		// Convert milliseconds to seconds.
		double secondsFadeout = _msFadeout / 1000.0;
		
		// 'fadeLenght' is the length of the fade in samples.
		// 'secondsFadeout' is a double, so '_sampleRate' is also casted to a double.
		unsigned int fadeLength = (unsigned int)( _sampleRate * secondsFadeout );
		assert( fadeLength );
		
		if ( _pos >= fadeLength )// Check if there are enough samples in the 'ringbuf' to apply the fadeout.
		{
			assert( fadeLength <= _size );// if 'fadeLength' is bigger than the size of '_bufferL' and '_bufferR' we have a problem.

			float step = 1.f / fadeLength;
			
			for ( unsigned int i = 0; i < fadeLength; i++ )
			{
				_bufferL[i] *= ( i * step );
				_bufferR[i] *= ( i * step );
			}
			//cout << "fadeLength:" << fadeLength << "\tstep:" << step << endl;
		}
		else// There are not enough samples in the 'ringbuf' to apply the fadeout.
		{
			// cout << "Not enought samples to 'applyFadeout()' !!" << endl;
			// A few ms of audio is not enough to do some serious scratching anyway, so I'll just clear the buffer.
			CRingBuffer::clearBuffer();
		}
	}
}

/*
When the ringBuffer is filled, '_pos' is the beginning of the audio. But not the beginning of the buffer. 
After calling 'unwindBuffer()', 'translateIndex'is not necessary anymore.
I hope this doesn't cause a cpu-usage-spike in the process block. 
Advantages: 
-Other member-functions of this class are easier to read without 'translateIndex()'.
-Other member-functions of this class perform better without 'translateIndex()'.

If the cpu-usage-spike is to big, 
an alternative implementation for this function would be to make _bufferL and _bufferR twice as big. 
Because then no memory would have to be allocated here. And only the samples before '_pos' would have to be copied.
But it will cost twice the memory.

Maybe I should rewrite this with for-loops for readability.
*/
void CRingBuffer::unwindBuffer()
{
	//cout << "CRingBuffer::unwindBuffer()" << endl;
	float * new_bufferL = new float[_size + 2]; // '+ 2' is for linear interpolation, cause we need 2 extra samples at the end..
	float * new_bufferR = new float[_size + 2];

	// Copy the samples after '_pos' to the beginning of the new buffer.
	for (unsigned int i = 0; i < (_size - _pos); i++)
	{
		new_bufferL[i] = _bufferL[i + _pos];
		new_bufferR[i] = _bufferR[i + _pos];
	}
	
	// Copy the samples before '_pos' to the end of the new buffer.
	for (unsigned int i = (_size - _pos); i < _size; i++)
	{
		unsigned int index = (i + _pos) % _size;

		new_bufferL[i] = _bufferL[ index ];
		new_bufferR[i] = _bufferR[ index ];
	}
	
	// Write 2 extra samples at the end for linear-interpolation.
	float lastSampleL = new_bufferL[_size - 1];
	float lastSampleR = new_bufferR[_size - 1];

	new_bufferL[_size] = lastSampleL;
	new_bufferR[_size] = lastSampleR;

	new_bufferL[_size + 1] = lastSampleL;
	new_bufferR[_size + 1] = lastSampleR;

	// delete[] old buffers.
	delete[] _bufferL;
	delete[] _bufferR;

	// new buffers are ready to be used.
	_bufferL = new_bufferL;
	_bufferR = new_bufferR;

	_pos = 0;
}

CRingBuffer::~CRingBuffer()
{
    deleteBuffer();
}
