#ifndef ringBuffer_H
#define ringBuffer_H

 #include <cassert>
#include <iostream>
#include <cmath>

using namespace std;


class CRingBuffer
{
    public:
		// Constructors
		CRingBuffer( CRingBuffer & ) = delete; // no accidental copy-construction.
		CRingBuffer();
		CRingBuffer( unsigned int bufferSize, unsigned int samplerate, float msFadeout );

		virtual ~CRingBuffer();

		void init( unsigned int ringbufSize, unsigned int samplerate, float msFadeout );
		
		// This class has 2 modes: Scratch-mode and fill mode. Fill-mode is default.

		// Since fill-mode is the default mode, we can start using 'processFill' to write samples to this ringbuf right away.
		// Write some samples to the ringbuffers. This should only be done in fill-mode.
		void processFill( float * L, float * R, unsigned int blockSize );

		// After 'processFill' there are samples in this ringbuffer.
		// After we prepared for scratching, we can start scratching.
		void prepareBufferForScratching();

		// Jeej !! :) Scratching :) 
		// 'indexes' is an array that contains the indexes to the samples in the ringbuffer.
		// The retreived samples will be written to 'outputBuf'
		void processScratch(float indexes[], unsigned int num_samples, float ** outputBuf);

		// When we are done scratching, 'prepareBufferForFilling' should be called. 
		// After that, all buffers are empty end reset. So we can start filling the buffers again.
		void prepareBufferForFilling()
		{
			CRingBuffer::reset();
		}

		// Get size of the ringbuffer.
		inline unsigned int size() const
		{
			return _size;
		}

		// Shows te values of the samples ( as in the array. indexes are not translated )
		// The result in fill-mode than the result in -scratch mode.
		// Because in 'prepereBufferForScratching()' de buffers are changed.
		void printBufferContent( string text = "" ) const;

		// Note !!: What about the extra sample at the end of the buffers for interpolation ?
		// These 2 can be made private, when I'm ready testing them.
		void reverseChannel_L();
		void reverseChannel_R();

		// Deletes old buffers and allocates new ones with size 'size'. 
		// And initializes elements to 0.f;
		// Old data is not copied to the new buffers.
		void setSize(unsigned int size);

		// delete[] buffers. (if there are any)
		void deleteBuffer();

//remove 'private:' for testing purposes. When testing this class I want to be able to access all members.
private:
        // Sets all samples to 0.f
		void clearBuffer();

		// Set state to initial condition. 
		// All elements in the ringbuf are set to 0.f. ('reset()' calles 'clearBuffer()'
		// Plus some other stuff.
		void reset();
		
		// Get 1 linear interpolated sample for each channel.
		void getSample( float index, float & L, float & R );

private:
		/* This should be called, when going from fill-mode to scratch mode.
		Checks if a fadeout is necessary, and applies it when it is.
		When the ringBuf wasn't entirely filled. It applies a fade-out at the end of the audio in the 'ringbuf',
		so we won't be scratchint over a click/glitch. */
		void applyFadeout();

		// Used for 'applyFadeout()'
		unsigned int _sampleRate;
		float _msFadeout;
		
		// These are used to determine if this RingBuffer is fully fillled. 
		// If it's not filled a fade-out will be applied.
		unsigned int _filledWithXSamples;
		// Is true when we have one full cycle of audio-samples. ( or more )
		bool _fullCycleFilled;

		/*
		'unwindBuffer()' SHOULD BE CALLED BEFORE WE'RE GOING TO SCRATCH !!
		When the ringBuffer is filled, '_pos' is the beginning audio. But not the beginning of the buffer. 
		I hope this doesn't cause a cpu-usage-spike in the process block. 
		Advantages: 
		-Other member-functions of this class are easier to read without 'translateIndex()'.
		-Other member-functions of this class perform better without 'translateIndex()'.
		*/
		void unwindBuffer();

		unsigned int _size;
		
		float * _bufferL;
		float * _bufferR;
        
        // Keeps track of where we are in the process of filling the ringbuffer.
		// '_pos' is the index of the sample after the last written sample.
		// 
		unsigned int _pos;
 };

#endif // ringBuffer_H
