#include "CScratch.h"


CScratch::CScratch()
	:
	_fadeOutTail( NULL )
{
	cout << "CScratch default constructor. Don't forget to call 'init' before use." << endl;
}

void CScratch::prepareToPlay(	int sampleRate,
								unsigned int audioRingbufSize,
								float msFadeTime,
								unsigned int guiWindowHeight,
								float paramSmoothLagTimeMs) 
{
	assert( sampleRate > 0 );
	assert( audioRingbufSize > 0 );
	assert( msFadeTime > 0.f );
	assert( guiWindowHeight > 0 );
	assert( paramSmoothLagTimeMs > 0.f );

	_dcFilterL.init( sampleRate );
	_dcFilterR.init( sampleRate );
	
	// We start out with the mouseButton up. aka fill-mode.
	_previousMode = Fill;
	
	// We start scrathing with pos = 1.f;
	_previousPos = 1.f;

	// Set the size of the audio ringbuffer. Maybe '_ringbuffer' and '_waveformData4GUI' should be aggregated in one object,
	// since they both contain a ringbuffer...
	_ringbuffer.init( audioRingbufSize, sampleRate, msFadeTime );

	// The number of samples, and the number of pixels they have to be converted to.
	_waveformData4GUIL.init( audioRingbufSize, guiWindowHeight / 2 );
	_waveformData4GUIR.init( audioRingbufSize, guiWindowHeight / 2 );
	
	_paramSmooth.init( paramSmoothLagTimeMs, (float) sampleRate /*samplerate*/ );
	// We start scratchPos parameter smoothing with input 1.f. So prepare for that.
	_paramSmooth.setZ( 1.f );

	// Fade-out tail x-fade stuff.
	_fadeOutTail = NULL;
	_numSamplesFadeTime = int( (float)sampleRate * ( msFadeTime / 1000.f ) );
	_fadeInStep = 1.f / (float)_numSamplesFadeTime;
	_fadeTimeCountDown = 0;
}

void CScratch::releaseResources()
{
	_ringbuffer.deleteBuffer();
	CScratch::deleteScratchFadeOutTail();
}

void CScratch::printBufferContent()
{
	_ringbuffer.printBufferContent();
}

void CScratch::createScratchFadeOutTail()
{
	// To avoid clicks caused by abrubtly switching from scratch-mode to fill-mode,
	// we need to fade-out the scratching audio, and fade-in the dry signal.
	// In order to create a fade-out for the scratching audio, we retreive some future samples.
	
	// If there currently is NO fadeOutTail, we have to create one.
	// If there is a fadeOutTail, it means we are busy fading, so we don't need a new fadeOutTail.
	// If '_ringbuffer' is filled with less than '_numSamplesFadeTime' samples, 
	// those samples are set to 0.f ...
	if (! _fadeOutTail )
	{
		_fadeTimeCountDown = _numSamplesFadeTime;

		_fadeOutTail = new float*[2];// Two channels. ( Why Dynamically allocate when the size is static anyway ? )
		_fadeOutTail[0] = new float[_numSamplesFadeTime]();
		_fadeOutTail[1] = new float[_numSamplesFadeTime]();

		// Retreive the future samples for the 'fadeOutTail'...
		CScratch::processScratchMode( _fadeOutTail, _numSamplesFadeTime, _previousPos/*or newPos ? */);

		// Do the fade-out: in place.
		float fadeStep = 1.f / (float)_numSamplesFadeTime;
		for ( int i = 0; i < _numSamplesFadeTime; i++ )
		{
			float factor = (_numSamplesFadeTime - i) * fadeStep;
			_fadeOutTail[0][i] *= factor;
			_fadeOutTail[1][i] *= factor;
		}
	}
}

// Add the fadeOutTail to the output, and fade in the dry signal.
void CScratch::processScratchFadeOutTail( float ** samples, int blockSize )
{
	//assert( _fadeTimeCountDown > 0 );
	//assert( _fadeOutTail );
	if ( _fadeOutTail )// if there is a fadeOutTail.
	{
		// How far are we in the process of x-fading ?
		int offset = _numSamplesFadeTime - _fadeTimeCountDown;

		// If there are enough fadeOut-samples to process a whole block of samples.
		int numSamples2Process = blockSize;

		// If there are NOT enough fadeOut-samples to process a whole block of samples.
		if ( _fadeTimeCountDown < blockSize ) numSamples2Process = _fadeTimeCountDown;

		for ( int i = 0; i < numSamples2Process; i++ )
		{
			// Fade-in the dry signal.
			// '_fadeInStep' = 1.f / _numSamplesFadeTime
			float fadeInFactor = _fadeInStep * (offset + i);
			samples[0][i] *= fadeInFactor;
			samples[1][i] *= fadeInFactor;

			// Add the scratchFadeOutTail to the output.
			samples[0][i] += _fadeOutTail[0][ offset + i ];
			samples[1][i] += _fadeOutTail[1][ offset + i ];
		}
		// The remaining (unprocessed) samples in the block are the unaffected dry signal.
		// When a whole block is processed there are no remaining samples.

		_fadeTimeCountDown -= numSamples2Process;

		assert( _fadeTimeCountDown >= 0 );

		// If we are done x-fading the fadeOutTail is not needed anymore.
		// '_fadeOutTail' is set to NULL, so we know we are not fading anymore.
		if( _fadeTimeCountDown == 0) CScratch::deleteScratchFadeOutTail();
	}
}

void CScratch::deleteScratchFadeOutTail()
{
	if ( _fadeOutTail )
	{
		delete[] _fadeOutTail[0];
		delete[] _fadeOutTail[1];
		delete[] _fadeOutTail;
		_fadeOutTail = NULL;
	}
}

void CScratch::prepareForFillMode()
{
	// CScratch::createScratchFadeOutTail();

	// Prepare for fill-mode.
	_ringbuffer.prepareBufferForFilling();
	_waveformData4GUIL.reset();
	_waveformData4GUIR.reset();
}

void CScratch::prepareForScratchMode()
{
	// Prepare for scratch-mode.
	_ringbuffer.prepareBufferForScratching();

	// Scratching always starts with pos 1.f. So prepare the paramSmooth filter for that.
	_paramSmooth.setZ( 1.f );

	// DC filter also has to be reset because we're going to feed it new audio.
	_dcFilterL.reset();
	_dcFilterR.reset();
}

// Asumes 2 channel audio.
// 'samples' is stereo input and output. 
// 'newPos' should be 1.f, 0.f or something inbetween.
// There are two modes: scratch-mode and fill-mode. 
// We start in fill-mode.
void CScratch::process ( float ** samples, unsigned int blockSize, float newPos, bool newMode )
{
	assert( newPos >= 0.f && newPos <= 1.f );

	if ( newMode == Scratch )// scratch-mode
	{
		if ( _previousMode == Mode::Fill )// First sample-block in scratch-mode. Prepare for scratching.
		{
			CScratch::prepareForScratchMode();
		}
		CScratch::processScratchMode( samples, blockSize, newPos );
		_previousPos = newPos;
	}
	else // fill-mode.
	{
		// Before 'process' is called, in 'prepareToPlay()',  '_previousMode' is initialized to Fill mode.
		if ( _previousMode == Mode::Scratch )// First sample-block in fill-mode. 
		{
			CScratch::createScratchFadeOutTail();
			CScratch::prepareForFillMode();
		}
		CScratch::processFillMode( samples, blockSize );
	}
	_previousMode = newMode;

	// 'CScratch::processScratchFadeOutTail' fades in the dry signal and adds the scratchFadeOutTail to the output.
	CScratch::processScratchFadeOutTail( samples, blockSize );
}

void CScratch::processFillMode( float ** samples, unsigned int blockSize )
{
	// Fill the ring buffers for the gui.
	_waveformData4GUIL.process( samples[0], blockSize );// Left channel.
	_waveformData4GUIR.process( samples[1], blockSize );// Right channel.
	
	// 'samples' is not modified. So the audio input is fed to the output.
	
	// Feed the incoming sample blocks to the '_ringbuffer'.
	_ringbuffer.processFill( samples[0], samples[1], blockSize );
}

void CScratch::processScratchMode(float ** samples, unsigned int blockSize, float newPos)
{
	// The scratching distance that has to be covered during this block.
	float blockDistance = newPos - _previousPos;

	// The scratching distance that has to be covered in 1 sample. Note!: step can also be negative...
	float step = blockDistance / (float)blockSize;

	// An array of indexes. These indexes are used to retreive samples from the ringbuffer.
	float * indexes = new float[blockSize];
	// float * indexesSmoothed = new float[blockSize];

	// Prepare 'indexes'
	for (unsigned int i = 0; i < blockSize; i++)
	{
		// Move from '_prevPos' to 'newPos' in 'blockSize' steps.
		// This creates a straight slope: From '_prevPos' to 'newPos' for this block of samples.
		indexes[i] = _previousPos + (step * (float)i);

		// Do some parameter smoothing on 'indexes' with 'CParamSmooth'.
		indexes[i] = _paramSmooth.process( indexes[i] );

		// At this point, an element in 'indexes' is 0.f, 1.f or something inbetween.
		// Multiply all 'indexes' elements with the size of the '_ringbuffer'. For use in '_ringbuffer.getSampleS()'.
		indexes[i] *= _ringbuffer.size();
		// The '_ringbuffer' actually has two extra elements at the end for interpolation.
		// ( internally the actual size is; '_size'+ 2 )
	}

	// Get samples from '_ringbuffer'. 'getSampleS()' writes the samples to 'samples'.
	_ringbuffer.processScratch( indexes, blockSize, samples );
	delete[] indexes;

	// Remove any DC offset.
	_dcFilterL.processBlock( samples[0], blockSize );
	_dcFilterR.processBlock( samples[1], blockSize );
}

CRingbufGUI & CScratch::getRingbufGUIL()
{
	return _waveformData4GUIL.getRingbufGUI();
}

CRingbufGUI & CScratch::getRingbufGUIR()
{
	return _waveformData4GUIR.getRingbufGUI();
}

CScratch::~CScratch()
{
	CScratch::deleteScratchFadeOutTail();
}


