#ifndef SCRATCH_H
#define SCRATCH_H

#include <iostream>

#include "CRingBuffer.h"
#include "CParamSmooth.h"
#include "CDCFilter.h"
#include "CWaveformData4GUI.h"

using namespace std;


class CScratch
{
    public:
		CScratch();// Default constructor. Don't forget to call 'prepareToPlay' before use.
		CScratch( CScratch & ) = delete; // no accidental copy-construction.
        
        virtual ~CScratch();

		void prepareToPlay(	int sampleRate,
							unsigned int audioRingbufSize, 
							float msFadeTime, 
							unsigned int guiWindowHeight, 
							float paramSmoothLagTimeMs );

		void releaseResources(); 
        
		// Assumes 2 channel audio. samples[0] = left channel. samples[1] = right channel.
		// 'samples' is a stereo input and stereo output.
		void process( float ** samples, unsigned int blockSize, float newPos, bool scratchMode );

		// A reference to a 'CRingbufGUI' is used by the GUI to draw the waveform.
		// 'WaveFormComponent' relies on 'CRingbufGUI' being there. 
		CRingbufGUI & getRingbufGUIL();
		CRingbufGUI & getRingbufGUIR();

		void printBufferContent();
	
	private:
		void prepareForFillMode();
		void prepareForScratchMode();

		// To store an amount of most recent audio-samples. in fill-mode, to scratch it later in scratch-mode.
		// The amount of recent samples that can be stored is set in 'prepareToPlay'.
		CRingBuffer _ringbuffer;

		// To store the last second of audio for the purpose of drawing it in the GUI.
		// This also contains a ringbuffer. 
		// This ringbuffer contains data that can be used to draw the waveform on a GUI.
		CWaveformData4GUI _waveformData4GUIL;
		CWaveformData4GUI _waveformData4GUIR; 
		
		// A lag-filter object to smooth-out 'scratchPosParameter' changes.
		CParamSmooth _paramSmooth;

		// Remove DC-offset. Use after: 'processScratchMode'
		CDCFilter _dcFilterL;
		CDCFilter _dcFilterR;

		// Scratch fade-out tail and x-fade stuff.
		int _numSamplesFadeTime;
		int _fadeTimeCountDown;
		float _fadeInStep;
		float ** _fadeOutTail;// Initially is NULL. 'createScratchFadeOutTail()' will assign an adress. 

		void createScratchFadeOutTail();
		
		// Used to create the x-fade. And deletes the fadeOutTail, when done fading.
		void processScratchFadeOutTail( float ** samples, int blockSize );
		
		// delete the fadeOutTail if there is one. if(_fadeOutTail != NULL )
		void deleteScratchFadeOutTail();
		//////////////////////////////////////////

		// The previous scratching position.
		float _previousPos;

		// The previous mouse-down state. True = mouse-down. False = mouse-up.
		bool _previousMode;

		enum Mode : bool {
			Scratch = true,
			Fill = false,
		};

		// The two modes. These are used in the public 'process' member function.
		void processFillMode( float ** samples, unsigned int blockSize );
		void processScratchMode( float ** samples, unsigned int blockSize, float newPos);
};

#endif // SCRATCH_H
