

#include "ScratchComponent.h"


//==============================================================================
ScratchComponent::ScratchComponent(	
					CRingbufGUI & ringbufL, 
					CRingbufGUI & ringbufR,
					AudioProcessorParameter & scratchPosParam,
					AudioProcessorParameter & mouseDownParam 
				)
	:
	waveFormL(ringbufL),
	waveFormR(ringbufR),
	scratchPosParam(scratchPosParam),
	mouseDownParam(mouseDownParam)
{
	timerIntervalMs = 1000 / 60; //60hz the waveForm will be drawn 60 times a second.
    
	halfWidth = Component::getWidth() / 2;
	halfHeight = Component::getHeight() / 2;

	// 'halfWidthPlus1' is to shift 'waveFormR' 1 pixel to the right.
	// So there is a line that separates the left and right channel, with the same colour of the background.
	halfWidthPlus1 = halfWidth + 1; 
	
	// x, y, width, height
	waveFormL.setBounds( 0, 0, halfWidth, halfHeight );
	waveFormR.setBounds( halfWidthPlus1, 0, halfWidth, halfHeight );

	Component::addAndMakeVisible( waveFormL );
	Component::addAndMakeVisible( waveFormR );
	
	// Start periodically drawing the waveForm.
	// 'ringbufL' and 'ringbufR' contain averaged sample values to be drawn periodically by this timer.
	ScratchComponent::Timer::startTimer( timerIntervalMs );

	playHead.setBounds( Component::getBounds() );
	Component::addAndMakeVisible( playHead );
}

// Is only called when the mouseButton is up.
void ScratchComponent::timerCallback()
{
	if ( mouseDownParam.getValue() )
	{
		int y = halfHeight - int(scratchPosParam.getValue() * halfHeight);

		// set position and paint waveForm on that position.
		waveFormL.setTopLeftPosition(0, y);
		waveFormR.setTopLeftPosition(halfWidthPlus1, y);
		
		// set colour and paint component.
		playHead.setPlayColour();
	}
	else
	{
		waveFormL.setTopLeftPosition(0, 0);
		waveFormR.setTopLeftPosition(halfWidthPlus1, 0);
		
		playHead.setRecColour();
	}
		
}

ScratchComponent::~ScratchComponent()
{
	ScratchComponent::Timer::stopTimer();
}

void ScratchComponent::paint (Graphics& g)
{
    g.fillAll (Colour (0xff001F36));
}

void ScratchComponent::resized()
{
    // This is called when the MainContentComponent is resized.
    // If you add any child components, this is where you should
    // update their positions.
	halfWidth = Component::getWidth() / 2;
	halfHeight = Component::getHeight() / 2;

	halfWidthPlus1 = halfWidth + 1;
	
	// x, y, width, height
	waveFormL.setBounds( 0, 0, halfWidth, halfHeight );
	waveFormR.setBounds( halfWidthPlus1, 0,  halfWidth, halfHeight );
	
	playHead.setBounds( Component::getBounds() );
}

void ScratchComponent::mouseDown(const MouseEvent& )
{
	waveFormL.setBufferedToImage(true);
	waveFormR.setBufferedToImage(true);

	// 'waveForm' isn't drawn periodically by the timer anymore. ( Only when waveForm is dragged.)
	ScratchComponent::Timer::stopTimer();
	
	// Change cursor to DraggingHandCursor. Cursor is changed back to 'NormalCursor' in 'mouseUp()'
	Component::setMouseCursor( MouseCursor::StandardCursorType::DraggingHandCursor );

	// Make 'playHead' green. This calls: 'playHead.repaint()'. 
	// Red means fillMode, and green means scratchMode:
	// green = play. red = rec.
	playHead.setPlayColour();

	// Let the host know we're going to change this parameter.
	mouseDownParam.beginChangeGesture();
	mouseDownParam.setValueNotifyingHost( 1.f );

	scratchPosParam.beginChangeGesture();
	scratchPosParam.setValueNotifyingHost(1.f);
}



void ScratchComponent::mouseDrag( const MouseEvent & event )
{
	int y = event.getDistanceFromDragStartY();
	
	// Set the position of the waveForms and paint them.
	waveFormL.setTopLeftPosition(0, y);
	waveFormR.setTopLeftPosition(halfWidthPlus1, y);

	// Keep 'y' in bounds. 
	if (y < 0) y = 0;
	else if (y > halfHeight) y = halfHeight;

	// Because 'y' is casted to a float, 'halfHeight' is also casted to a float.
	scratchPosParam.setValueNotifyingHost(1.f - (float(y) / halfHeight));
}

void ScratchComponent::mouseUp( const MouseEvent& )
{
	// Make the 'waveFormL' and 'waveFormR' draw 'CRingbufGUI' again.
	// Also to respond to automation.
	ScratchComponent::Timer::startTimer( timerIntervalMs );
	
	// We are not dragging anymore so change the mouse-cursor back to normal.
	Component::setMouseCursor( MouseCursor::StandardCursorType::NormalCursor );

	waveFormL.setTopLeftPosition( 0, 0 );
	waveFormR.setTopLeftPosition( halfWidthPlus1, 0 );

	playHead.setRecColour();//make playHead red. This calls: playHead.repaint()

	waveFormL.setBufferedToImage( false );
	waveFormR.setBufferedToImage( false );

	// Jump back to pos 1.f instantly ?
	scratchPosParam.setValueNotifyingHost(1.f);
	scratchPosParam.endChangeGesture();

	mouseDownParam.setValueNotifyingHost(0.f);
	mouseDownParam.endChangeGesture();
}
